!>\brief    Contrast transfer functions for electron micrographs
module ContrastTransferFunctions
    use Globals
    use Units
    implicit none
    private
    public :: ContrastTransferFunction

    type :: ContrastTransferFunction
        private
        real    ::  spherical_aberration                =   0.0             !<  Spherical aberration
        integer ::  spherical_aberration_units          =   millimeters
        real    ::  wavelength                          =   0.0             !<  Wavelength of the electrons
        integer ::  wavelength_units                    =   angstroms
        real    ::  amplitude_contrast                  =   0.0             !<  Fraction of amplitude contrast (from 0. to 1., usually btw .07 and .14, see Mindell 03)
        real    ::  defocus_1                           =   0.0             !< Underfocus along first axis  (positive values for underfocus; larger value = more underfocus)
        real    ::  defocus_2                           =   0.0             !< Underfocus along second axis
        integer ::  defocus_1_units                     =   microns
        integer ::  defocus_2_units                     =   microns
        real    ::  astigmatism_azimuth                 =   0.0             !< Azimuth of first axis. 0.0 means axis is at 3 o'clock
        integer ::  astigmatism_azimuth_units           =   degrees
        ! Fitting parameters
        real    ::  lowest_frequency_for_fitting        =   0.0
        integer ::  lowest_frequency_for_fitting_units  =   reciprocal_pixels
        real    ::  highest_frequency_for_fitting       =   0.5
        integer ::  highest_frequency_for_fitting_units =   reciprocal_pixels
        real    ::  astigmatism_tolerance               =   0.0                 !<  Expected (tolerated) astigmatism. During parameter search, astigmatism values much larger than this will be penalised against. Set to 0.0 to ignore this restraint.
        integer ::  astigmatism_tolerance_units         =   angstroms
        contains
            procedure,  public      ::  Init
            procedure,  public      ::  PrintInfo
            procedure,  public      ::  EvaluateAtSquaredSpatialFrequency
            procedure               ::  SetDefocusScalars
            procedure               ::  SetDefocusArray
            generic,    public      ::  SetDefocus => SetDefocusArray, SetDefocusScalars
            procedure,  public      ::  HasPixelUnits
            procedure,  public      ::  GetLowestFrequencyForFitting
            procedure,  public      ::  GetHighestFrequencyForFitting
            procedure,  public      ::  GetAstigmatismTolerance
            procedure,  public      ::  GetAstigmatism
            procedure,  public      ::  GetDefocusParametersInAngstromsAndDegrees
            procedure,  public      ::  SetDefocusParametersInAngstromsAndDegrees
            procedure,  public      ::  GetDefocusParameters
            procedure               ::  ConvertToPixelsAndRadians
            procedure,  public      ::  GetDefocus1InAngstroms
            procedure,  public      ::  GetDefocus2InAngstroms
            procedure,  public      ::  GetAstigmatismAzimuthInDegrees
    end type

    contains

    !>  \brief  Initialise a CTF object
    subroutine Init(self,acceleration_voltage,spherical_aberration,amplitude_contrast, &
                         defocus_1,defocus_2,astigmatism_azimuth, &
                         lowest_frequency_for_fitting,highest_frequency_for_fitting, &
                         astigmatism_tolerance, &
                         pixel_size)
        ! arguments
        class(ContrastTransferFunction),    intent(inout)   ::  self
        real,                               intent(in)      ::  acceleration_voltage            !<  Kilovolts
        real,                               intent(in)      ::  spherical_aberration            !<  mm
        real,                               intent(in)      ::  amplitude_contrast              !<  fraction
        real,                               intent(in)      ::  defocus_1                       !<  um
        real,                               intent(in)      ::  defocus_2                       !<  um
        real,                               intent(in)      ::  astigmatism_azimuth             !<  degrees
        real,                               intent(in)      ::  lowest_frequency_for_fitting    !<  1/A
        real,                               intent(in)      ::  highest_frequency_for_fitting   !<  1/A
        real,                               intent(in)      ::  astigmatism_tolerance           !<  A
        real,                               intent(in)      ::  pixel_size                      !<  A
        ! start work
        self%wavelength = akv_to_wl(acceleration_voltage)
        self%wavelength_units = angstroms
        self%spherical_aberration = spherical_aberration
        self%spherical_aberration_units = millimeters
        self%amplitude_contrast = amplitude_contrast
        self%defocus_1 = defocus_1
        self%defocus_2 = defocus_2
        self%defocus_1_units = microns
        self%defocus_2_units = microns
        self%astigmatism_azimuth = astigmatism_azimuth
        !if (present(lowest_frequency_for_fitting)) then
            self%lowest_frequency_for_fitting = lowest_frequency_for_fitting
            self%lowest_frequency_for_fitting_units = reciprocal_angstroms
        !endif
        !if (present(highest_frequency_for_fitting)) then
            self%highest_frequency_for_fitting = highest_frequency_for_fitting
            self%highest_frequency_for_fitting_units = reciprocal_angstroms
        !endif
        !if (present(astigmatism_tolerance)) then
            self%astigmatism_tolerance = astigmatism_tolerance
            self%astigmatism_tolerance_units = angstroms
        !endif
        !if (present(pixel_size)) then
            call self%ConvertToPixelsAndRadians(pixel_size)
        !endif
    end subroutine Init

    !>  \brief  Set defocus and astigmatism
    subroutine SetDefocusScalars(self,defocus_1,defocus_2,astigmatism_azimuth)
        ! arguments
        class(ContrastTransferFunction),    intent(inout)   ::  self
        real,                               intent(in)      ::  defocus_1               !<  pixels
        real,                               intent(in)      ::  defocus_2               !<  pixels
        real,                               intent(in)      ::  astigmatism_azimuth     !<  radians
        !
        self%defocus_1 = defocus_1
        self%defocus_1_units = pixels
        self%defocus_2 = defocus_2
        self%defocus_2_units = pixels
        self%astigmatism_azimuth = astigmatism_azimuth
        self%astigmatism_azimuth_units = radians
    end subroutine SetDefocusScalars
    subroutine SetDefocusArray(self,my_array)
        class(ContrastTransferFunction),    intent(inout)   ::  self
        real,                               intent(in)      ::  my_array(:)
        call self%SetDefocusScalars(my_array(1),my_array(2),my_array(3))
    end subroutine SetDefocusArray
    subroutine SetDefocusParametersInAngstromsAndDegrees(self,defocus_parameters,pixel_size)
        class(ContrastTransferFunction),    intent(inout)   ::  self
        real,                               intent(in)      ::  defocus_parameters(3)
        real,                               intent(in)      ::  pixel_size
        self%defocus_1 = defocus_parameters(1) / pixel_size
        self%defocus_1_units = pixels
        self%defocus_2 = defocus_parameters(2) / pixel_size
        self%defocus_2_units = pixels
        self%astigmatism_azimuth = convert(defocus_parameters(3),degrees,radians)
        self%astigmatism_azimuth_units = radians
    end subroutine

    function GetDefocusParameters(self) result(defocus)
        class(ContrastTransferFunction),    intent(in)      ::  self
        real                                                ::  defocus(3)
        defocus = [self%defocus_1,self%defocus_2,self%astigmatism_azimuth]
    end function
    function GetDefocusParametersInAngstromsAndDegrees(self,pixel_size) result (defocus)
        class(ContrastTransferFunction),    intent(in)      ::  self
        real                                                ::  defocus(3)
        real,                               intent(in)      ::  pixel_size
        defocus(1) = convert(self%defocus_1,self%defocus_1_units,angstroms,pixel_size)
        defocus(2) = convert(self%defocus_2,self%defocus_2_units,angstroms,pixel_size)
        defocus(3) = convert(self%astigmatism_azimuth,self%astigmatism_azimuth_units,degrees)
    end function
    real function GetDefocus1InAngstroms(self,pixel_size)
        class(ContrastTransferFunction),    intent(in)  ::  self
        real,   optional,                   intent(in)  ::  pixel_size
        GetDefocus1InAngstroms = convert(self%defocus_1,self%defocus_1_units,angstroms,pixel_size)
    end function
    real function GetDefocus2InAngstroms(self, pixel_size)
        class(ContrastTransferFunction),    intent(in)  ::  self
        real,   optional,                   intent(in)  ::  pixel_size
        GetDefocus2InAngstroms = convert(self%defocus_2,self%defocus_2_units,angstroms,pixel_size)
    end function
    real function GetAstigmatismAzimuthInDegrees(self) result(azimuth)
        class(ContrastTransferFunction),    intent(in)  ::  self
        azimuth = convert(self%astigmatism_azimuth,self%astigmatism_azimuth_units,degrees)
        azimuth = azimuth - 180.0e0 * nint(azimuth/180.0e0)
    end function

    subroutine PrintInfo(self)
        ! arguments
        class(ContrastTransferFunction),    intent(in)      ::  self
        ! start work
        write(*,'(a)')              '** ContrastTransferFunction **'
        write(*,'(a,f0.3,1x,a)')    'Wavelength = ', self%wavelength, unit_to_string(self%wavelength_units)
        write(*,'(a,f0.3,1x,a)')    'Spherical aberration = ', self%spherical_aberration, &
                                                                unit_to_string(self%spherical_aberration_units)
        write(*,'(a,f0.3)')         'Amplitude contrast = ', self%amplitude_contrast
        write(*,'(a,f0.3,1x,a)')    'Defocus 1 = ', self%defocus_1, unit_to_string(self%defocus_1_units)
        write(*,'(a,f0.3,1x,a)')    'Defocus 2 = ', self%defocus_2, unit_to_string(self%defocus_2_units)
        write(*,'(a,f0.3,1x,a)')    'Azimuth of astigmatism = ', self%astigmatism_azimuth, &
                                                                unit_to_string(self%astigmatism_azimuth_units)
        write(*,'(2(a,f0.3,a))')    'Frequencies for fitting: from ', self%lowest_frequency_for_fitting, &
                                    unit_to_string(self%lowest_frequency_for_fitting_units), ' to ', &
                                    self%highest_frequency_for_fitting, unit_to_string(self%highest_frequency_for_fitting_units)
        write(*,'(a,f0.1,1x,a)')    'Astigmatism tolerance = ', self%astigmatism_tolerance, &
                                                                unit_to_string(self%astigmatism_tolerance_units)
    end subroutine PrintInfo

    !>  \brief  Check that all units are in pixels / radians
    pure logical function HasPixelUnits(self)
        class(ContrastTransferFunction),    intent(in)      ::  self
        HasPixelUnits =         self%spherical_aberration_units == pixels &
                        .and.   self%wavelength_units == pixels &
                        .and.   self%defocus_1_units == pixels &
                        .and.   self%defocus_2_units == pixels &
                        .and.   self%astigmatism_azimuth_units == radians &
                        .and.   self%highest_frequency_for_fitting_units == reciprocal_pixels &
                        .and.   self%lowest_frequency_for_fitting_units == reciprocal_pixels &
                        .and.   self%astigmatism_tolerance_units == pixels
    end function HasPixelUnits

    pure real function GetLowestFrequencyForFitting(self) result(lowest_frequency)
        class(ContrastTransferFunction),    intent(in)  ::  self
        lowest_frequency = self%lowest_frequency_for_fitting
    end function

    pure real function GetHighestFrequencyForFitting(self) result(highest_frequency)
        class(ContrastTransferFunction),    intent(in)  ::  self
        highest_frequency = self%highest_frequency_for_fitting
    end function

    pure real function GetAstigmatismTolerance(self) result(astigmatism_tolerance)
        class(ContrastTransferFunction),    intent(in)  ::  self
        astigmatism_tolerance = self%astigmatism_tolerance
    end function

    pure real function GetAstigmatism(self) result(astigmatism)
        class(ContrastTransferFunction),    intent(in)  ::  self
        astigmatism = self%defocus_1 - self%defocus_2
    end function


    !>  \brief  Convert all values to pixel/radians units
    subroutine ConvertToPixelsAndRadians(self,pixel_size)
        class(ContrastTransferFunction),    intent(inout)   ::  self
        real,                               intent(in)      ::  pixel_size
        ! start work
        call unit_conversion(self%spherical_aberration,self%spherical_aberration_units,pixels,pixel_size)
        call unit_conversion(self%wavelength,self%wavelength_units,pixels,pixel_size)
        call unit_conversion(self%defocus_1,self%defocus_1_units,pixels,pixel_size)
        call unit_conversion(self%defocus_2,self%defocus_2_units,pixels,pixel_size)
        call unit_conversion(self%astigmatism_azimuth,self%astigmatism_azimuth_units,radians)
        call unit_conversion(self%lowest_frequency_for_fitting,self%lowest_frequency_for_fitting_units, &
                            reciprocal_pixels,pixel_size)
        call unit_conversion(self%highest_frequency_for_fitting,self%highest_frequency_for_fitting_units, &
                            reciprocal_pixels,pixel_size)
        call unit_conversion(self%astigmatism_tolerance,self%astigmatism_tolerance_units,pixels,pixel_size)
    end subroutine ConvertToPixelsAndRadians

    !>  \brief wrapper around eval_ctf which takes a ctf object as argument
    pure elemental function EvaluateAtSquaredSpatialFrequency(self,squared_spatial_frequency,azimuth, &
                                                                return_sign_only,pixel_size) result(ctf)
        ! arguments
        class(ContrastTransferFunction),intent(in)      ::  self
        real,                           intent(in)      ::  squared_spatial_frequency   !<  squared reciprocal pixels (0.25 is Nyquist)
        real,                           intent(in)      ::  azimuth                     !<  radians. azimuth at which to evalulate the ctf
        logical,        optional,       intent(in)      ::  return_sign_only            !<  return the sign of the ctf rather than its value
        real,           optional,       intent(in)      ::  pixel_size                  !<  Needed if CTF not already in correct units
        !
        real(kind=4)                                    ::  ctf
        !
        ctf      = eval_ctf_slave(  self%spherical_aberration,self%wavelength,self%amplitude_contrast,    &
                                    self%defocus_1,self%defocus_2,self%astigmatism_azimuth, &
                                    squared_spatial_frequency,azimuth,return_sign_only)
    end function EvaluateAtSquaredSpatialFrequency

    !>  \brief returns the ctf, based on ctffind3 subroutine (see mindell 2003)
    pure elemental real function eval_ctf_slave(cs,wl,ampl_cont,dfmid1,dfmid2,angast,spa_freq_sq,ang,sign_only)
        implicit none
        ! arguments
        real,                   intent(in)  ::  cs          !< spherical aberation (pixels)
        real,                   intent(in)  ::  wl          !< electron wavelength (pixels)
        real,                   intent(in)  ::  ampl_cont   !< fraction of amplitude contrast (from .07 to .14, see mindell 2003)
        real,                   intent(in)  ::  dfmid1      !< defocus along first axis (pixels)
        real,                   intent(in)  ::  dfmid2      !< defocus along second axis (for astigmatic ctf, dfmid1 .ne. dfmid2) (pixels)
        real,                   intent(in)  ::  angast      !< azimuth of first axis. 0.0 means axis is at 3 o'clock. (radians)
        real,                   intent(in)  ::  spa_freq_sq !< squared spatial frequency at which to compute the ctf (1/pixels^2)
        real,                   intent(in)  ::  ang         !< angle at which to compute the ctf (radians)
        logical,    optional,   intent(in)  ::  sign_only   !<  return 1.0 or -1.0
        ! private variables
        real    ::  wgh1    ! phase contrast
        real    ::  wgh2    ! amplitude contrast
        real    ::  ctf_arg
        ! start work

        ! compute phase and amplitude contrast
        wgh1    =   sqrt(1-ampl_cont**2)
        wgh2    =   ampl_cont

        ctf_arg = eval_ctf_arg(cs,wl,dfmid1,dfmid2,angast,spa_freq_sq,ang)
        eval_ctf_slave = - wgh1 * sin(ctf_arg) - wgh2 * cos(ctf_arg)

        if (present(sign_only)) then
            if (sign_only) eval_ctf_slave = sign(1.0,eval_ctf_slave)
        endif
    end function eval_ctf_slave



    !>  \brief returns the argument (radians) to the sine and cosine terms of the ctf
    pure elemental real function eval_ctf_arg(cs,wl,dfmid1,dfmid2,angast,spa_freq_sq,ang) result(ctf_arg)
        implicit none
        ! arguments
        real,                   intent(in)  ::  cs          !< spherical aberation (pixels)
        real,                   intent(in)  ::  wl          !< electron wavelength (pixels)
        real,                   intent(in)  ::  dfmid1      !< defocus along first axis (pixels)
        real,                   intent(in)  ::  dfmid2      !< defocus along second axis (for astigmatic ctf, dfmid1 .ne. dfmid2) (pixels)
        real,                   intent(in)  ::  angast      !< azimuth of first axis. 0.0 means axis is at 3 o'clock. (radians)
        real,                   intent(in)  ::  spa_freq_sq !< square of spatial frequency at which to compute the ctf (1/pixels^2)
        real,                   intent(in)  ::  ang         !< angle at which to compute the ctf (radians)
        ! private variables
        real    ::  df      ! defocus at point at which we're evaluating the ctf
        real    ::  ccos
        ! start work
        ! compute the defocus
        ccos = cos(2.0*(ang-angast))
        df = 0.5*(dfmid1+dfmid2+ccos*(dfmid1-dfmid2))
        ! compute the ctf argument
        ctf_arg = pi * wl * spa_freq_sq * (df - 0.5 * wl**2 * spa_freq_sq * cs)
    end function eval_ctf_arg

    !>  \brief returns the argument (radians) to the sine and cosine terms of the ctf
    pure subroutine eval_ctf_arg_array(cs,wl,dfmid1,dfmid2,angast,spa_freq_sq,ang,ctf_arg)
        implicit none
        ! arguments
        real,                   intent(in)  ::  cs              !< spherical aberation (pixels)
        real,                   intent(in)  ::  wl              !< electron wavelength (pixels)
        real,                   intent(in)  ::  dfmid1          !< defocus along first axis (pixels)
        real,                   intent(in)  ::  dfmid2          !< defocus along second axis (for astigmatic ctf, dfmid1 .ne. dfmid2) (pixels)
        real,                   intent(in)  ::  angast          !< azimuth of first axis. 0.0 means axis is at 3 o'clock. (radians)
        real,                   intent(in)  ::  spa_freq_sq(:)  !< square of spatial frequency at which to compute the ctf (1/pixels^2)
        real,                   intent(in)  ::  ang             !< angle at which to compute the ctf (radians)
        real,                   intent(out) ::  ctf_arg(:)
        ! private variables
        real    ::  df      ! defocus at point at which we're evaluating the ctf
        real    ::  ccos
        ! start work
        ! compute the defocus
        ccos = cos(2.0*(ang-angast))
        df = 0.5*(dfmid1+dfmid2+ccos*(dfmid1-dfmid2))
        ! compute the ctf argument
        ctf_arg(:) = pi * wl * spa_freq_sq * (df - 0.5 * wl**2 * spa_freq_sq(:) * cs)
    end subroutine eval_ctf_arg_array

    !>  \brief  convert acceleration voltage in kv into electron wavelength in angstroms
    pure real function akv_to_wl(akv)
        implicit none
        real, intent(in) :: akv !<  kV
        akv_to_wl = 12.26/sqrt(1000.0*akv+0.9784*(1000.0*akv)**2/(10.0**6.0))
    end function akv_to_wl


end module
